from trading_bots.bots import Bot


class {{ pascal_case_bot_name }}(Bot):
    label = '{{ bot_name }}'

    def _setup(self, config):
        # Setup your Bot before execution
        pass

    def _algorithm(self):
        # Implement your Bot's logic here
        pass

    def _abort(self):
        # Abort logic, runs on exception
        pass
